﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using ArcGIS.Core.CIM;
using ArcGIS.Core.Geometry;
using ArcGIS.Desktop.Mapping;
using orbitgt.mapping3d.sdk.viewer;

namespace plugin_3dm_arcgispro
{
    /// <summary>
    /// Class MapObjectOverlay displays a map object as overlay on the map view.
    /// </summary>
    class MapObjectOverlay
    {
        /** The map object */
        private MapObject mapObject;
        /** The map view */
        private MapView mapView;

        /** The geometry */
        private Geometry geometry;
        /** The geometry symbol */
        private CIMSymbolReference geometrySymbol;
        /** The geometry overlay handle */
        private System.IDisposable geometryOverlay;
        /** The recycled geometry overlay from an previous map overlay object */
        private System.IDisposable recycledGeometryOverlay;

        /** The label graphic */
        private CIMTextGraphic labelGraphic;
        /** The label overlay handle */
        private System.IDisposable labelOverlay;

        /// <summary>
        /// Create a new map overlay.
        /// </summary>
        /// <param name="mapObject">the map object to display</param>
        /// <param name="mapView">the map view to display on</param>
        /// <param name="addToMapView">should the map object be added to the map view?</param>
        /// <param name="recycledGeometryOverlay">the optional recycled geometry overlay</param>
        internal MapObjectOverlay(MapObject mapObject, MapView mapView, bool addToMapView, System.IDisposable recycledGeometryOverlay)
        {
            /* Initialize */
            this.mapObject = mapObject;
            this.mapView = mapView;
            this.geometry = null;
            this.geometrySymbol = null;
            this.geometryOverlay = null;
            this.recycledGeometryOverlay = recycledGeometryOverlay;
            this.labelGraphic = null;
            this.labelOverlay = null;
            /* Area object? */
            if (mapObject.spatialArea != null) CreateArea();
            /* Line object? */
            else if (mapObject.spatialLine != null) CreateLine();
            /* Point object? */
            else if (mapObject.spatialPoint != null && mapObject.symbolPath == null) CreatePoint();
            /* Symbol object? */
            else if (mapObject.spatialPoint != null && mapObject.symbolPath != null) CreateSymbol();
            /* Label? */
            if (mapObject.spatialPoint != null && mapObject.label != null) CreateLabel();
            /* Add to the map view? */
            if (addToMapView) AddToMapView();
        }

        /// <summary>
        /// Close the map overlay
        /// </summary>
        /// <param name="recycleGeometryOverlay">if true recycles the geometry overlay if there is one</param>
        /// <returns>an optional geometry overlay to recycle</returns>
        internal System.IDisposable Close(bool recycleGeometryOverlay)
        {
            if (this.labelOverlay != null)
            {
                this.labelOverlay.Dispose();
                this.labelOverlay = null;
            }
            if (this.geometryOverlay != null)
            {
                if (recycleGeometryOverlay) return this.geometryOverlay;
                this.geometryOverlay.Dispose();
                this.geometryOverlay = null;
            }
            return null;
        }

        /// <summary>
        /// Convert a 0..255 alpha value into a CIM color alpha value
        /// </summary>
        /// <param name="alpha">the 0..255 alpha value</param>
        /// <returns>the CIM color alpha value</returns>
        static private double ConvertAlpha(double alpha)
        {
            return alpha / 255.0 * 100.0;
        }

        /// <summary>
        /// Create an overlay for an area object
        /// </summary>
        private void CreateArea()
        {
            /* Create the geometry */
            this.geometry = MapObjects.GetMapObjectGeometry(mapObject, mapView);
            if (this.geometry == null) return;
            /* Create the symbol */
            CIMColor lineColor = ColorFactory.Instance.CreateRGBColor(mapObject.color.red, mapObject.color.green, mapObject.color.blue, mapObject.color.alpha);
            CIMColor fillColor = ColorFactory.Instance.CreateRGBColor(mapObject.fillColor.red, mapObject.fillColor.green, mapObject.fillColor.blue, ConvertAlpha(mapObject.fillColor.alpha));
            CIMStroke areaOutline = SymbolFactory.Instance.ConstructStroke(lineColor, mapObject.lineWidth, SimpleLineStyle.Solid);
            CIMPolygonSymbol areaSymbol = SymbolFactory.Instance.ConstructPolygonSymbol(fillColor, SimpleFillStyle.Solid, areaOutline);
            this.geometrySymbol = areaSymbol.MakeSymbolReference();
        }

        /// <summary>
        /// Create an overlay for a line object
        /// </summary>
        private void CreateLine()
        {
            /* Create the geometry */
            this.geometry = MapObjects.GetMapObjectGeometry(mapObject, mapView);
            if (this.geometry == null) return;
            /* Create the symbol */
            CIMColor lineColor = ColorFactory.Instance.CreateRGBColor(mapObject.color.red, mapObject.color.green, mapObject.color.blue, ConvertAlpha(mapObject.color.alpha));
            CIMStroke outline = SymbolFactory.Instance.ConstructStroke(lineColor, mapObject.lineWidth, SimpleLineStyle.Solid);
            CIMLineSymbol lineSymbol = SymbolFactory.Instance.ConstructLineSymbol(outline);
            this.geometrySymbol = lineSymbol.MakeSymbolReference();
        }

        /// <summary>
        /// Create an overlay for a point object
        /// </summary>
        private void CreatePoint()
        {
            /* Create the geometry */
            this.geometry = MapObjects.GetMapObjectGeometry(mapObject, mapView);
            if (this.geometry == null) return;
            /* Get the properties */
            double symbolSize = (mapObject.symbolSize / 18.0 * 12.0); // a point symbol of size 12 is drawn as 18 pixels on screen
            double lineWidth = (mapObject.lineWidth / 18.0 * 12.0);
            CIMColor lineColor = ColorFactory.Instance.CreateRGBColor(mapObject.color.red, mapObject.color.green, mapObject.color.blue, ConvertAlpha(mapObject.color.alpha));
            CIMColor fillColor = ColorFactory.Instance.CreateRGBColor(mapObject.fillColor.red, mapObject.fillColor.green, mapObject.fillColor.blue, ConvertAlpha(mapObject.fillColor.alpha));
            /* Create the symbol */
            CIMPointSymbol pointSymbol = SymbolFactory.Instance.ConstructPointSymbol(fillColor, symbolSize, SimpleMarkerStyle.Circle);
            CIMVectorMarker vectorMarker = pointSymbol.SymbolLayers[0] as CIMVectorMarker;
            CIMPolygonSymbol polygonSymbol = vectorMarker.MarkerGraphics[0].Symbol as CIMPolygonSymbol;
            polygonSymbol.SymbolLayers[0] = SymbolFactory.Instance.ConstructStroke(lineColor, lineWidth, SimpleLineStyle.Solid); //This is the outline
            polygonSymbol.SymbolLayers[1] = SymbolFactory.Instance.ConstructSolidFill(fillColor); //This is the fill
            this.geometrySymbol = pointSymbol.MakeSymbolReference();
        }

        /// <summary>
        /// Create an overlay for a point symbol
        /// </summary>
        private void CreateSymbol()
        {
            /* Create the geometry */
            this.geometry = MapObjects.GetSymbolPathGeometry(mapObject, mapView);
            if (this.geometry == null) return;
            /* Create the (area) symbol */
            CIMColor lineColor = ColorFactory.Instance.CreateRGBColor(mapObject.color.red, mapObject.color.green, mapObject.color.blue, mapObject.color.alpha);
            CIMColor fillColor = ColorFactory.Instance.CreateRGBColor(mapObject.fillColor.red, mapObject.fillColor.green, mapObject.fillColor.blue, ConvertAlpha(mapObject.fillColor.alpha));
            CIMStroke areaOutline = SymbolFactory.Instance.ConstructStroke(lineColor, mapObject.lineWidth, SimpleLineStyle.Solid);
            CIMPolygonSymbol areaSymbol = SymbolFactory.Instance.ConstructPolygonSymbol(fillColor, SimpleFillStyle.Solid, areaOutline);
            this.geometrySymbol = areaSymbol.MakeSymbolReference();
        }

        /// <summary>
        /// Create an overlay for a text label
        /// </summary>
        private void CreateLabel()
        {
            /* Get the label properties */
            Coordinate labelPosition = mapObject.labelPosition;
            if (labelPosition == null) labelPosition = Coordinate.fromXY("pixel", 0.0, 0.0);
            Color labelColor = mapObject.labelColor;
            if (labelColor == null) labelColor = new Color(255, 255, 255, 255);
            double labelFontSize = mapObject.labelFontSize;
            if (labelFontSize < 1.0) labelFontSize = 12;
            /* Create the symbol */
            Coordinate mapOrigin = MapObjects.GetMapObjectCentroid(mapObject);
            if (mapOrigin == null) return;
            MapPoint mapLabelPosition = MapPointBuilder.CreateMapPoint(mapOrigin.x, mapOrigin.y, mapOrigin.z, MapObjects.VIEWER_CRS);
            mapLabelPosition = GeometryEngine.Instance.Project(mapLabelPosition, MapObjects.HOST_CRS) as MapPoint;
            CIMColor mapLabelColor = ColorFactory.Instance.CreateRGBColor(labelColor.red, labelColor.green, labelColor.blue, ConvertAlpha(labelColor.alpha));
            CIMTextSymbol labelSymbol = SymbolFactory.Instance.ConstructTextSymbol(mapLabelColor, labelFontSize);
            double textScale = 72.0 / 96.0; // text offset is 72 dpi, mapobject offset is 96 dpi
            //textScale *= (24.0 / System.Math.Sqrt(labelPosition.x * labelPosition.x + labelPosition.y * labelPosition.y)); // put at 24 pixel distance?
            labelSymbol.OffsetX = textScale * labelPosition.x;
            labelSymbol.OffsetY = -textScale * labelPosition.y; // text offset goes bottom-up
            labelSymbol.HorizontalAlignment = HorizontalAlignment.Center;
            labelSymbol.VerticalAlignment = VerticalAlignment.Center;
            /* Create the graphic */
            this.labelGraphic = new CIMTextGraphic();
            this.labelGraphic.Shape = mapLabelPosition;
            this.labelGraphic.Symbol = labelSymbol.MakeSymbolReference();
            this.labelGraphic.Text = mapObject.label;
        }

        /// <summary>
        /// Add the map object to the map view
        /// </summary>
        internal void AddToMapView()
        {
            if (geometry!=null)
            {
                if (this.geometryOverlay == null)
                {
                    if (this.recycledGeometryOverlay == null)
                    {
                        this.geometryOverlay = mapView.AddOverlay(geometry, geometrySymbol);
                    }
                    else
                    {
                        this.geometryOverlay = this.recycledGeometryOverlay;
                        this.recycledGeometryOverlay = null;
                        mapView.UpdateOverlay(this.geometryOverlay, geometry, geometrySymbol);
                    }
                }
            }
            if (labelGraphic != null)
            {
                if (this.labelOverlay==null) this.labelOverlay = mapView.AddOverlay(labelGraphic); // UpdateOverlay of a CIMGraphic is not yet possible
            }
        }

        /// <summary>
        /// Remove the map object from the map view
        /// </summary>
        internal void RemoveFromMapView()
        {
            if (this.geometryOverlay != null)
            {
                this.geometryOverlay.Dispose();
                this.geometryOverlay = null;
            }
            if (this.labelOverlay != null)
            {
                this.labelOverlay.Dispose();
                this.labelOverlay = null;
            }
        }
    }
}
