﻿using System;
using System.Diagnostics;
using ArcGIS.Core.Data;
using ArcGIS.Core.Geometry;
using ArcGIS.Desktop.Mapping;
using com.orbitgis.tool.message;
using orbitgt.mapping3d.sdk.viewer;

namespace plugin_3dm_arcgispro
{
    /// <summary>
    /// Class Overlays manages the overlay of ArcGIS features on the 3D Mapping viewer
    /// </summary>
    class Overlays
    {
        // define the (outline) color of the overlays
        static private Color COLOR = Color.fromRGBA(0, 0, 255, 255);
        // define the fill color of the overlays
        static private Color FILL_COLOR = Color.fromRGBA(0, 0, 255, 128);
        // define the outline line width
        static private double LINE_WIDTH = 1.0;

        // Convert a segment list to a part
        //
        static private CoordinatePart segmentListToPart(string crs, ReadOnlySegmentCollection path,bool hasZ)
        {
            CoordinatePart part = new CoordinatePart();
            part.crs = crs;
            for (int i=0; i<path.Count; i++)
            {
                Segment segment = path[i];
                MapPoint startPoint = segment.StartPoint;
                Coordinate startCoordinate;
                if (hasZ) startCoordinate = Coordinate.fromXYZ(crs, startPoint.X, startPoint.Y, startPoint.Z);
                else startCoordinate = Coordinate.fromXY(crs, startPoint.X, startPoint.Y);
                MapPoint endPoint = segment.EndPoint;
                Coordinate endCoordinate;
                if (hasZ) endCoordinate = Coordinate.fromXYZ(crs, endPoint.X, endPoint.Y, endPoint.Z);
                else endCoordinate = Coordinate.fromXY(crs, endPoint.X, endPoint.Y);
                if (i == 0) part.points.add(startCoordinate);
                part.points.add(endCoordinate);
            }
            return part;
        }

        // Create new map objects from a feature
        //
        static private void createMapObjects(FeatureLayer featureLayer,Feature feature,string layerName, AList<MapObject> mapObjects)
        {
            Geometry geometry = feature.GetShape();
            if (geometry==null)
            {
                Message.WriteLine("Feature has no geometry");
                return;
            }
            SpatialReference layerSR = featureLayer.GetSpatialReference();
            if (layerSR == null)
            {
                Message.WriteLine("Layer has no spatial reference");
                return;
            }
            string crs = "" + layerSR.Wkid;
            bool hasZ = geometry.HasZ;
            String objectId = featureLayer.Name + "#" + feature.GetObjectID();
            // do we have a point?
            if (geometry is MapPoint)
            {
                MapPoint point = (MapPoint)geometry;
                Message.WriteLine("Found point (CRS "+ crs+")");
                MapObject mapObject = new MapObject();
                if (hasZ) mapObject.spatialPoint = Coordinate.fromXYZ(crs, point.X, point.Y, point.Z);
                else mapObject.spatialPoint = Coordinate.fromXY(crs, point.X, point.Y);
                mapObject.color = COLOR;
                mapObject.objectId = objectId;
                mapObjects.add(mapObject);
                return;
            }
            // do we have a multi-point?
            if (geometry is Multipoint)
            {
                Multipoint mpoint = (Multipoint)geometry;
                Message.WriteLine("Found multipoint (CRS " + crs + ")");
                for (int i=0; i<mpoint.PointCount; i++)
                {
                    MapPoint point = mpoint.Points[i];
                    MapObject mapObject = new MapObject();
                    if (hasZ) mapObject.spatialPoint = Coordinate.fromXYZ(crs, point.X, point.Y, point.Z);
                    else mapObject.spatialPoint = Coordinate.fromXY(crs, point.X, point.Y);
                    mapObject.color = COLOR;
                    mapObject.objectId = objectId + "#" + i;
                    mapObjects.add(mapObject);
                }
                return;
            }
            // do we have a polyline?
            if (geometry is Polyline)
            {
                Polyline polyline = (Polyline)geometry;
                Message.WriteLine("Found polyline (CRS " + crs + ")");
                CoordinatePath path = new CoordinatePath();
                path.crs = crs;
                foreach(ReadOnlySegmentCollection part in polyline.Parts)
                {
                    CoordinatePart cpart = segmentListToPart(crs, part, hasZ);
                    if (cpart != null) path.parts.add(cpart);
                }
                MapObject mapObject = new MapObject();
                mapObject.spatialLine = path;
                mapObject.color = COLOR;
                mapObject.lineWidth = LINE_WIDTH;
                mapObject.objectId = objectId;
                mapObjects.add(mapObject);
                return;
            }
            // do we have a polygon?
            if (geometry is Polygon)
            {
                Polygon polygon = (Polygon)geometry;
                Message.WriteLine("Found polygon (CRS " + crs + ")");
                CoordinatePath path = new CoordinatePath();
                path.crs = crs;
                foreach (ReadOnlySegmentCollection part in polygon.Parts)
                {
                    CoordinatePart cpart = segmentListToPart(crs, part, hasZ);
                    if (cpart != null) path.parts.add(cpart);
                }
                MapObject mapObject = new MapObject();
                mapObject.spatialArea = path;
                mapObject.color = COLOR;
                mapObject.fillColor = FILL_COLOR;
                mapObject.lineWidth = LINE_WIDTH;
                mapObject.objectId = objectId;
                mapObjects.add(mapObject);
                return;
            }
            Message.WriteLine("Feature not added as overlay, the geometry class is: " + geometry.GetType().Name);
            return;
        }

        /// <summary>
        /// Add the selected ArcGIS features as overlays on the 3D Mapping viewer
        /// </summary>
        /// <param name="mapView"></param>
        /// <param name="viewer"></param>
        static internal void doAddOverlays(MapView mapView, Viewer3DM viewer)
        {
            try
            {
                // create a map object for each selected feature
                Message.WriteLine("Adding the selected features as overlays");
                AList<MapObject> mapObjects = new AList<MapObject>(MapObject.TYPE);
                foreach (MapMember layer in mapView.Map.GetSelection().Keys)
                {
                    // we should have a vector layer
                    if (layer is FeatureLayer)
                    {
                        FeatureLayer featureLayer = (FeatureLayer)layer;
                        RowCursor selectedFeatures = featureLayer.GetSelection().Search();
                        while (selectedFeatures.MoveNext())
                        {
                            Feature feature = (Feature)selectedFeatures.Current;
                            Message.WriteLine("Adding selected feature "+feature.GetObjectID()+" of '"+featureLayer.Name+"'");
                            createMapObjects(featureLayer, feature, featureLayer.Name, mapObjects);
                        }
                    }
                }
                Message.WriteLine("Created " + mapObjects.size() + " map objects");
                // send the map objects to the viewer
                if (mapObjects.size() > 0) viewer.addOverlayObjects(mapObjects);
                else viewer.clearOverlayObjects();
                // clear the feature selection to indicate we have processed the objects
                mapView.Map.ClearSelection();
            }
            catch (Exception e)
            {
                /* Log the exception */
                Message.WriteLine("Error while adding overlays: " + e);
            }
        }
    }
}
