﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Timers;
using ArcGIS.Core.Geometry;
using ArcGIS.Desktop.Framework;
using ArcGIS.Desktop.Framework.Threading.Tasks;
using ArcGIS.Desktop.Mapping;
using ArcGIS.Desktop.Mapping.Events;
using com.orbitgis.tool.message;
using orbitgt.mapping3d.sdk.viewer;
using OrbitGT.Mapping3D.SDK;

namespace plugin_3dm_arcgispro
{
    /// <summary>
    /// Class PluginArcGIS implements the 3D Mapping Viewer plugin for ArcGIS Pro
    /// </summary>
    class PluginArcGIS : SDKViewerApplication3DM
    {
        /** The single instance of this class */
        static internal PluginArcGIS INSTANCE = null;

        private SDKViewerHost viewerHost;
        private MapObjectLayer footprintLayer;
        private MapObjectLayer fieldOfViewLayer;
        private MapObjectLayer measureLayer;
        private MapObjectLayer globalLayer;
        private UpdateFootprint footprintUpdater;
        private UpdateFieldOfView fieldOfViewUpdater;
        private UpdateMeasurement measurementUpdater;
        private UpdateGlobal globalUpdater;
        private Timer refreshTimer;

        // the map view that shows the viewer overlays
        private MapView mapView0;

        /// <summary>
        /// Create a new plugin instance
        /// </summary>
        public PluginArcGIS()
        {
            try
            {
                // create a logfile
                string userDir = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData), "OrbitGT");
                if (Directory.Exists(userDir)==false) Directory.CreateDirectory(userDir);
                string logFileName = Path.Combine(userDir, "arcgispro_3dmc_log.txt");
                Message.setLogFile(new StreamWriter(logFileName));
                Message.WriteLine("Creating ArcGIS plugin");
                Message.WriteLine("Logging to file '" + logFileName + "'");
                Message.WriteLine("Start time "+DateTime.Now.ToString("yyyy.MM.dd HH:mm:ss"));
                Message.WriteLine("Machine name is '" + System.Environment.MachineName + "'");
                // initialize the mapping system
                MapView mapView = this.getMapView();
                MapObjects.VIEWER_CRS = SpatialReferenceBuilder.CreateSpatialReference(MapObjects.VIEWER_CRS_CODE);
                MapObjects.HOST_CRS = SpatialReferenceBuilder.CreateSpatialReference(MapObjects.HOST_CRS_CODE);
                // initialize the 3D Mapping viewer (starts the viewer host and the websocket server)
                // we pass the auto-login application state (remember-me) and coordinate system preference using the startup options
                this.viewerHost = new SDKViewerHost();
                AMap<object> viewerOptions = new AMap<object>(Collections.TYPE_OBJECT);
                string appState = this.readApplicationStateFile();
                if (appState != null) viewerOptions.set(orbitgt.mapping3d.sdk.viewer.Constants.STARTUP_APP_STATE, appState);
                viewerOptions.set(orbitgt.mapping3d.sdk.viewer.Constants.STARTUP_CRS, MapObjects.CRS);
                Message.WriteLine("Initializing viewer");
                this.viewerHost.InitializeViewer(this, viewerOptions);
                Message.WriteLine("Initialized viewer");
                // listen to various signals emitted by the 3D Mapping viewer
                this.getViewer().onAppStateChanged.SignalEvent += OnAppStateChanged;
                this.getViewer().onFootprintsChanged.SignalEvent += OnFootprintsChanged;
                this.getViewer().onFieldOfViewsChanged.SignalEvent += OnFieldOfViewsChanged;
                this.getViewer().onMeasureStateChanged.SignalEvent += OnMeasureStateChanged;
                this.getViewer().onGlobalsChanged.SignalEvent += OnGlobalsChanged;
                this.getViewer().onZoomToExtent.SignalEvent += OnZoomToExtent;
                this.getViewer().onPickPositionRequested.SignalEvent += OnPickPositionRequested;
                this.getViewer().onPickPositionCancelled.SignalEvent += OnPickPositionCancelled;
                // start the viewer (which opens a javascript browser window that connects to this plugin instance using websockets)
                Message.WriteLine("Starting viewer");
                this.viewerHost.StartViewer();
                Message.WriteLine("Started viewer");
                // create the layers
                Message.WriteLine("Creating map layers");
                this.footprintLayer = new MapObjectLayer("3D Mapping Footprints", this);
                this.fieldOfViewLayer = new MapObjectLayer("3D Mapping Field of Views", this);
                this.measureLayer = new MapObjectLayer("3D Mapping Measurements", this);
                this.globalLayer = new MapObjectLayer("3D Mapping Globals", this);
                // create the updaters for the layers
                Message.WriteLine("Creating map updaters");
                this.footprintUpdater = new UpdateFootprint(this.footprintLayer, this);
                this.fieldOfViewUpdater = new UpdateFieldOfView(this.fieldOfViewLayer, this);
                this.measurementUpdater = new UpdateMeasurement(this.measureLayer, this);
                this.globalUpdater = new UpdateGlobal(this.measureLayer, this);
                // start the refresh timer
                Message.WriteLine("Creating timer");
                this.refreshTimer = new Timer(250); // milliseconds
                this.refreshTimer.Elapsed += doRefresh;
                this.refreshTimer.AutoReset = true;
                this.refreshTimer.Enabled = true;
                // listen to changes in the mapcanvas extent
                Message.WriteLine("Adding mapview listener");
                MapViewCameraChangedEvent.Subscribe(this.updatedMapCanvas);
                // done
                Message.WriteLine("Created ArcGIS plugin");
            }
            catch (Exception e)
            {
                /* Log the exception */
                Message.WriteLine("Exception while creating ArcGIS plugin: " + e);
            }
        }

        /// <summary>
        /// Get the map view.
        /// </summary>
        /// <returns></returns>
        internal MapView getMapView()
        {
            if (this.mapView0==null)
            {
                this.mapView0 = MapView.Active;
                if (this.mapView0 == null) Message.WriteLine("Failed to find an active mapview!");
                else Message.WriteLine("Found an active mapview");
            }
            return this.mapView0;
        }

        /// <summary>
        /// Close the plugin
        /// </summary>
        public void Close()
        {
            this.refreshTimer.Stop();
        }

        /// <summary>
        /// get the name of the application
        /// </summary>
        /// <returns>a descriptive name</returns>
        string SDKViewerApplication3DM.GetApplicationName()
        {
            return "ArcGIS Pro Plugin by Orbit GT";
        }

        /// <summary>
        /// Check if the plugin is ready for use
        /// </summary>
        /// <returns></returns>
        public bool isReady()
        {
            return this.viewerHost.isViewerReady();
        }

        /// <summary>
        /// Get the 3D Mapping viewer
        /// </summary>
        /// <returns></returns>
        public Viewer3DM getViewer()
        {
            return this.viewerHost.GetViewer();
        }

        /// <summary>
        /// Read the application state file from disk
        /// </summary>
        /// <returns></returns>
        private string readApplicationStateFile()
        {
            string userDir = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData), "OrbitGT");
            string fileName = Path.Combine(userDir, "arcgispro_3dmc_app_state.ini");
            if (File.Exists(fileName) == false) return null;
            long stateSize = new FileInfo(fileName).Length;
            if (stateSize == 0) return null;
            string state = "";
            FileStream reader = new FileStream(fileName, FileMode.Open);
            for (int i = 0; i < stateSize; i++) state += (char)reader.ReadByte();
            reader.Close();
            if (state != null) Message.WriteLine("Read app state (size " + state.Length + ") from '" + fileName + "'");
            return state;
        }

        /// <summary>
        /// Write the application state file to disk
        /// </summary>
        /// <param name="appState"></param>
        private void writeApplicationStateFile(string appState)
        {
            if (appState == null) return;
            string userDir = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData), "OrbitGT");
            if (Directory.Exists(userDir) == false) Directory.CreateDirectory(userDir);
            string fileName = Path.Combine(userDir, "arcgispro_3dmc_app_state.ini");
            Message.WriteLine("Writing app state (size " + appState.Length + ") to '" + fileName + "'");
            FileStream writer = new FileStream(fileName, FileMode.Create);
            for (int i = 0; i < appState.Length; i++) writer.WriteByte((byte)appState[i]);
            writer.Close();
        }

        /// <summary>
        /// Signal called by the 3D Mapping viewer when the application (login) state has changed. We save the new state to disk to enable auto-login (remember-me) for the next day
        /// </summary>
        /// <param name="value"></param>
        private void OnAppStateChanged(string appState)
        {
            Message.WriteLine("Got 'onAppStateChanged' signal (" + appState + ")");
            this.writeApplicationStateFile(appState);
        }
        /// <summary>
        /// Signal called by the 3D Mapping viewer the footprints have changed
        /// </summary>
        /// <param name="value"></param>
        private void OnFootprintsChanged(string message)
        {
            Message.WriteLine("Got 'onFootprintsChanged' signal");
            // request new footprints
            this.footprintUpdater.updatedMapCanvas();
        }
        /// <summary>
        /// Signal called by the 3D Mapping viewer when the field-of-views have changed
        /// </summary>
        /// <param name="value"></param>
        private void OnFieldOfViewsChanged(string message)
        {
            Message.WriteLine("Got 'onFieldOfViewsChanged' signal");
            // request new field-of-views
            this.fieldOfViewUpdater.updatedViews();
        }
        /// <summary>
        /// Signal called by the 3D Mapping viewer when the measure state has changed
        /// </summary>
        /// <param name="value"></param>
        private void OnMeasureStateChanged(string message)
        {
            Message.WriteLine("Got 'onMeasureStateChanged' signal");
            // request a new measure state
            this.measurementUpdater.updatedMeasurement();
        }
        /// <summary>
        /// Signal called by the 3D Mapping viewer when the global map objects have changed
        /// </summary>
        /// <param name="value"></param>
        private void OnGlobalsChanged(string message)
        {
            Message.WriteLine("Got 'onGlobalsChanged' signal");
            // request new globals
            this.globalUpdater.updatedGlobals();
        }
        /// <summary>
        /// Signal called by the 3D Mapping viewer when a zoomout to a spatial extent is requested by the user (zoom-to-extent step 1)
        /// </summary>
        /// <param name="value"></param>
        private void OnZoomToExtent(Bounds bounds)
        {
            Message.WriteLine("Got 'onZoomToExtent' signal (" + bounds + ")");
            this.OnZoomToExtent2(bounds);
        }
        /// <summary>
        /// Do the zoom-to-extent procedure (zoom-to-extent step 2)
        /// </summary>
        /// <param name="bounds">bounds the bounds to zoom the map to</param>
        private void OnZoomToExtent2(Bounds bounds)
        {
            MapView mapView = this.getMapView();
            if (mapView == null) return;
            if (bounds.valid == false) return;
            QueuedTask.Run(() =>
            {
                // create a polygon of the bounds
                PolygonBuilder builder = new PolygonBuilder(MapObjects.VIEWER_CRS);
                List<Coordinate2D> coordinates = new List<Coordinate2D>();
                coordinates.Add(new Coordinate2D(bounds.minX, bounds.minY));
                coordinates.Add(new Coordinate2D(bounds.maxX, bounds.minY));
                coordinates.Add(new Coordinate2D(bounds.maxX, bounds.maxY));
                coordinates.Add(new Coordinate2D(bounds.minX, bounds.maxY));
                builder.AddPart(coordinates);
                Polygon boundsArea = builder.ToGeometry();
                // zoom the map view
                mapView.ZoomTo(boundsArea, null/*duration*/, true/*maintainViewDirection*/);
            });
        }
        /// <summary>
        /// Signal called by the 3D Mapping viewer when the viewer needs a location to be picked on the reference map by the user (pick-position step 1)
        /// </summary>
        /// <param name="value"></param>
        private void OnPickPositionRequested(string message)
        {
            Message.WriteLine("Got 'onPickPositionRequested' signal");
            this.OnPickPositionRequested2();
        }
        /// <summary>
        /// Activate the pick-position tool (pick-position step 2)
        /// </summary>
        async private void OnPickPositionRequested2()
        {
            // activate the tool
            await FrameworkApplication.SetCurrentToolAsync(ToolPickLocation.DAML_ID);
            // switch the tool to "pick-position" mode
            if (ToolPickLocation.INSTANCE != null) ToolPickLocation.INSTANCE.pickPositionMode = true;
        }
        /// <summary>
        /// Signal called by the 3D Mapping viewer when the pick-position operation needs to be cancelled (cancel pick-position step 1)
        /// </summary>
        /// <param name="value"></param>
        private void OnPickPositionCancelled(string value)
        {
            Message.WriteLine("Got 'onPickPositionCancelled' signal");
        }

        /// <summary>
        /// Call this method to request an update to the map (can be called from a non-UI thread).
        /// </summary>
        /// <param name="reason">the reason to redraw the map</param>
        internal void RequestMapUpdate(string reason)
        {
            Message.WriteLine("Requesting map redraw for reason '" + reason + "'");
            if (this.viewerHost == null) return;
            if (this.viewerHost.isViewerReady() == false) return;
            MapView mapView = this.getMapView();
            if (mapView == null) return;
            QueuedTask.Run(() =>
            {
                try
                {
                    // update all layers
                    this.footprintLayer.updateMapView(mapView);
                    this.fieldOfViewLayer.updateMapView(mapView);
                    this.measureLayer.updateMapView(mapView);
                    this.globalLayer.updateMapView(mapView);
                }
                catch (Exception e)
                {
                    /* Log the exception */
                    Message.WriteLine("Exception during map update: " + e);
                }
            });
        }

        /// <summary>
        /// Called by ArcGIS when the mapcanvas extent has changed (from MapViewCameraChangedEvent)
        /// </summary>
        /// <param name="obj"></param>
        private void updatedMapCanvas(MapViewCameraChangedEventArgs obj)
        {
            MapView mapView = this.getMapView();
            if (mapView == null) return;
            if (obj.MapView == mapView)
            {
                // request new footprints and field-of-views
                this.footprintUpdater.updatedMapCanvas();
                this.fieldOfViewUpdater.updatedViews();
            }
        }

        // Called by the refresh timer at a regular interval
        //
        static private void doRefresh(Object source, ElapsedEventArgs e)
        {
            PluginArcGIS plugin = PluginArcGIS.INSTANCE;
            if ((plugin!=null) && (plugin.viewerHost.isViewerReady()))
            {
                plugin.footprintUpdater.checkForUpdates();
                plugin.fieldOfViewUpdater.checkForUpdates();
                plugin.measurementUpdater.checkForUpdates();
                plugin.globalUpdater.checkForUpdates();
            }
        }

        /// <summary>
        /// Called when the user pressed the 'start' button on the toolbar again
        /// </summary>
        internal void restart()
        {
            // start a new viewer (which opens a javascript browser window that connects to this plugin instance using websockets)
            if (this.viewerHost.hasViewerConnection() == false) this.viewerHost.StartViewer();
            // update all map objects
            this.footprintUpdater.updatedMapCanvas();
            this.fieldOfViewUpdater.updatedViews();
            this.measurementUpdater.updatedMeasurement();
            this.globalUpdater.updatedGlobals();
        }

        /// <summary>
        /// Toolbar action callback method
        /// </summary>
        /// <param name="visible"></param>
        internal void doShowFootprints(bool visible)
        {
            if (this.viewerHost.isViewerReady())
            {
                this.footprintLayer.setVisible(visible);
                this.fieldOfViewLayer.setVisible(visible);
                this.measureLayer.setVisible(visible);
                this.globalLayer.setVisible(visible);
                this.RequestMapUpdate("toggled footprint visibility: "+ visible);
            }
        }

        /// <summary>
        /// Toolbar action callback method
        /// </summary>
        internal void doSaveMeasurement()
        {
            try
            {
                Message.WriteLine("doSaveMeasurement");
                MapView mapView = this.getMapView();
                if (mapView == null) return;
                if (this.viewerHost.isViewerReady())
                {
                    this.measurementUpdater.doSaveMeasurement(mapView);
                }
            }
            catch (Exception e)
            {
                /* Log the exception */
                Message.WriteLine("Exception while saving measurement: " + e);
            }
        }

        /// <summary>
        /// Toolbar action callback method
        /// </summary>
        internal void doAddOverlays()
        {
            try
            {
                Message.WriteLine("doAddOverlays");
                MapView mapView = this.getMapView();
                if (mapView == null) return;
                if (this.viewerHost.isViewerReady())
                {
                    Overlays.doAddOverlays(mapView, this.getViewer());
                }
            }
            catch (Exception e)
            {
                /* Log the exception */
                Message.WriteLine("Exception while adding overlays: " + e);
            }
        }
    }
}
